import io
import shutil
import sys
from pathlib import Path
from typing import Any
from urllib.request import urlopen
from zipfile import ZipFile

from hatchling.builders.hooks.plugin.interface import BuildHookInterface
from tqdm import tqdm

_VERSION = '20221217'
_URL = ('https://github.com/openslide/openslide-winbuild/releases/download'
        f'/v{_VERSION}/openslide-win64-{_VERSION}.zip')
_TARGET = 'src/bipl/io/libs'


def _url_to_io(url: str) -> io.BytesIO:
    r = urlopen(url)

    total = int(value) if (value := r.headers['content-length']) else None
    buf = io.BytesIO()
    with tqdm.wrapattr(r, 'read', total, desc='retrieve libraries') as fp:
        shutil.copyfileobj(fp, buf)
    return buf


def _download_dlls(folder: Path) -> None:
    """Download archive, extract DLLs and place them into target folder"""
    if sys.platform != 'win32' or folder.exists():
        return

    folder.mkdir(parents=True, exist_ok=True)
    try:
        with ZipFile(_url_to_io(_URL)) as zf:
            for zfpath in zf.namelist():
                if not zfpath.endswith('.dll'):
                    continue
                with zf.open(zfpath) as src, \
                        (folder / Path(zfpath).name).open('wb') as dst:
                    shutil.copyfileobj(src, dst)
    except BaseException:
        for p in folder.glob('*.dll'):
            p.unlink()
        folder.rmdir()
        raise


class CustomBuildHook(BuildHookInterface):
    def initialize(self, version: str, build_data: dict[str, Any]) -> None:
        if self.target_name != 'wheel':
            return

        build_data |= {
            'tag': 'py3-none-win_amd64',  # lock platform
            'pure_python': False,  # got DLLs inside
        }
        _download_dlls(Path(self.root, _TARGET))



from __future__ import annotations

__all__ = ['Slide']

import os
from bisect import bisect_right
from pathlib import Path
from typing import final

import cv2
import numpy as np
from glow import memoize

from ._openslide import Openslide
from ._slide_bases import REGISTRY, Driver, Item, Lod, normalize
from ._tiff import Tiff

# TODO: inside Slide.open import ._slide.registry,
# TODO: and in ._slide.registry do registration and DLL loading
# TODO: to make Slide export not require DLL presence

Openslide.register('bif mrxs ndpi scn svs svsslide tif tiff vms vmu')
Tiff.register('svs tif tiff')

_MAX_BYTES = int(os.environ.get('GLOW_SLIDE_BYTES') or 102_400)


@memoize(capacity=_MAX_BYTES, policy='lru')
def _cached_open(path: Path) -> Slide:
    if not path.exists():
        raise FileNotFoundError(path)

    if tps := REGISTRY.get(path.suffix):
        last_exc = BaseException()
        for tp in reversed(tps):  # Loop over types to find non-failing
            try:
                return Slide(path, tp)
            except ValueError as exc:
                last_exc = exc
        raise last_exc from None
    raise ValueError(f'Unknown file format {path}')


def _fit_to(image: np.ndarray, dsize: tuple[int, ...]):
    if image.shape[:2] == dsize:
        return image
    return cv2.resize(image, dsize[::-1], interpolation=cv2.INTER_AREA)


@final
class Slide:
    """Usage:
    ```
    slide = Slide.open('test.svs')
    shape: tuple[int, ...] = slide.shape
    scales: tuple[int, ...] = slide.scales

    # Get numpy.ndarray
    image: np.ndarray = slide[:2048, :2048]
    ```
    """
    # TODO: check if memory leak
    # TODO: add __enter__/__exit__/close to make memory management explicit
    # TODO: call .close in finalizer
    path: Path
    shape: tuple[int, ...]
    spacing: float | None
    pools: tuple[int, ...]
    lods: tuple[Lod, ...]
    extras: dict[str, Item]

    def __init__(self, path: Path, driver_cls: type[Driver]):
        self.path = path
        driver = driver_cls(path)

        num_items = len(driver)
        assert num_items > 0

        item_0, *items = (driver[idx] for idx in range(num_items))
        assert isinstance(item_0, Lod)

        lods: dict[int, Lod] = {1: item_0}
        self.extras: dict[str, Item] = {}
        for item in items:
            if isinstance(item, Lod):
                pool = round(item_0.shape[0] / item.shape[0])
                lods[pool] = item
            elif key := item.key:
                self.extras[key] = item

        self.pools = *sorted(lods.keys()),
        self.lods = *(lods[pool] for pool in self.pools),
        self.shape = self.lods[0].shape
        self.spacing = self.lods[0].spacing

        self.extras |= driver.named_items()
        self.driver = driver_cls.__name__

    def __reduce__(self) -> tuple:
        return Slide.open, (self.path, )

    def __repr__(self) -> str:
        line = f"'{self.path}', shape={self.shape}, pools={self.pools}"
        if self.spacing:
            line += f', spacing={self.spacing}'
        line += f', driver={self.driver}'
        return f'{type(self).__name__}({line})'

    def best_lod_for(self, pool: float) -> tuple[int, Lod]:
        """Gives the most detailed LOD below `pool`"""
        idx = max(bisect_right(self.pools, pool) - 1, 0)
        return self.pools[idx], self.lods[idx]

    def pool(self, zoom: int) -> Lod:
        """Use like `slide.pool(4)[y0:y1, x0:x1]` call"""
        p, lod = self.best_lod_for(zoom)
        if p == zoom:
            return lod
        assert zoom % p == 0, 'fractional pooling is not supported'
        return lod.downscale(zoom // p)

    def __getitem__(self, key: slice | tuple[slice, ...]) -> np.ndarray:
        """Retrieves tile"""
        # TODO: Ignore step, always redirect to self.lods[0].__getitem__
        slices = normalize(key, self.shape)

        step0, step1 = (s.step for s in slices)
        if step0 != step1:
            raise ValueError('slice steps should be the same for each axis')
        if step0 <= 0:
            raise ValueError('slice steps should be positive')

        pool, lod = self.best_lod_for(step0)
        slices = *(slice(s.start // pool, s.stop // pool) for s in slices),
        image = lod.crop(slices)

        ratio = pool / step0
        dsize = *(round(ratio * s) for s in image.shape[:2]),
        return _fit_to(image, dsize)

    def at(self,
           z0_yx_offset: tuple[int, ...],
           dsize: int | tuple[int, ...],
           scale: float = 1) -> np.ndarray:
        """Read square region starting with offset"""
        dsize = dsize if isinstance(dsize, tuple) else (dsize, dsize)
        assert len(dsize) == 2

        pool, lod = self.best_lod_for(scale)
        slices = *(slice(int(c) // pool,
                         int(c + d * scale) // pool)
                   for c, d in zip(z0_yx_offset, dsize)),
        image = lod.crop(slices)
        return _fit_to(image, dsize)

    def extra(self, name: str) -> np.ndarray | None:
        if item := self.extras.get(name):
            return np.array(item, copy=False, order='C')
        return None

    def thumbnail(self) -> np.ndarray:
        item = self.extras.get('thumbnail')
        if not item:
            item = self.lods[-1]
        return np.asarray(item)

    @classmethod
    def open(cls, anypath: Path | str) -> Slide:
        """Open multi-scale image."""
        path = Path(anypath).resolve().absolute()
        return _cached_open(path)



from __future__ import annotations

__all__ = ['Driver', 'Item', 'Lod', 'REGISTRY']

from dataclasses import dataclass
from pathlib import Path
from typing import final

import cv2
import numpy as np

REGISTRY: dict[str, list[type[Driver]]] = {}


def normalize(slices: tuple[slice, ...] | slice,
              shape: tuple[int, ...]) -> tuple[slice, ...]:
    """Ensures slices to be exactly 2 slice with non-none endpoints"""
    if isinstance(slices, slice):
        slices = slices, slice(None)
    assert len(slices) == 2
    return *(slice(
        s.start if s.start is not None else 0,
        s.stop if s.stop is not None else axis_len,
        s.step if s.step is not None else 1,
    ) for s, axis_len in zip(slices, shape)),


@dataclass(frozen=True)
class Item:
    shape: tuple[int, ...]

    @property
    def key(self) -> str | None:
        raise NotImplementedError

    def __array__(self) -> np.ndarray:
        raise NotImplementedError


@dataclass(frozen=True)
class Lod(Item):
    spacing: float | None

    @final
    @property
    def key(self) -> None:
        return None

    def crop(self, slices: tuple[slice, ...]) -> np.ndarray:
        """Reads crop of LOD. Overridable"""
        raise NotImplementedError

    @final
    def __getitem__(self, key: slice | tuple[slice, ...]) -> np.ndarray:
        """Reads crop of LOD"""
        slices = normalize(key, self.shape)
        assert all(s.step == 1 for s in slices)
        return self.crop(slices)

    @final
    def __array__(self) -> np.ndarray:
        """Reads whole LOD in single call"""
        return self[:, :]

    def downscale(self, pool: int) -> Lod:
        if pool == 1:
            return self
        h, w, c = self.shape
        return ProxyLod(
            (h // pool, w // pool, c),
            (self.spacing * pool if self.spacing else None),
            pool,
            self.base if isinstance(self, ProxyLod) else self,
        )


@dataclass(frozen=True)
class ProxyLod(Lod):
    pool: int
    base: Lod

    def crop(self, slices: tuple[slice, ...]) -> np.ndarray:
        src_slices = *(slice(s.start * self.pool, s.stop * self.pool)
                       for s in slices),
        image = self.base[src_slices]

        shape = *((s.stop - s.start) for s in slices),
        return cv2.resize(image, shape[::-1], interpolation=cv2.INTER_AREA)


class Driver:
    @final
    @classmethod
    def register(cls, extensions: str):
        """Registers type builder for extensions. Last call takes precedence"""
        for ext in extensions.split():
            REGISTRY.setdefault(f'.{ext}', []).append(cls)

    def __init__(self, path: Path) -> None:
        raise NotImplementedError

    def __len__(self) -> int:
        """Count of indexed items"""
        return 0

    def __getitem__(self, index: int) -> Item:
        """Gives indexed item"""
        raise NotImplementedError

    def named_items(self) -> dict[str, Item]:
        keys = self.keys()
        return {k: self.get(k) for k in keys}

    def keys(self) -> list[str]:
        """List of names for named items"""
        return []

    def get(self, key: str) -> Item:
        """Gives named item"""
        raise NotImplementedError



"""
Driver based on OpenSlide
- slow
- thread-safe
- compatible with formats: tiff/tif/svs, ndpi/vms/vmu, scn, mrxs, svsslide, bif
"""
# TODO: handle viewport offsets
from __future__ import annotations

__all__ = ['Openslide']

import weakref
from collections.abc import Iterator
from ctypes import (POINTER, addressof, byref, c_char_p, c_double, c_int64,
                    c_ubyte, c_void_p)
from dataclasses import dataclass
from pathlib import Path

import cv2
import numpy as np

from ._libs import load_library
from ._slide_bases import Driver, Item, Lod

OSD = load_library('libopenslide', 0)

OSD.openslide_open.restype = POINTER(c_ubyte)
OSD.openslide_get_error.restype \
    = OSD.openslide_get_property_value.restype \
    = c_char_p

OSD.openslide_get_property_names.restype \
    = OSD.openslide_get_associated_image_names.restype \
    = POINTER(c_char_p)

OSD.openslide_get_level_downsample.restype = c_double


def _ntas_to_iter(null_terminated_array_of_strings) -> Iterator[bytes]:
    if not null_terminated_array_of_strings:
        return
    i = 0
    while s := null_terminated_array_of_strings[i]:
        yield s
        i += 1


_LUT_CA = (np.outer(np.arange(256), np.arange(256)) // 255).astype('u1')


def _mbgra_to_rgb(bgra: np.ndarray, rgb_base: np.ndarray) -> np.ndarray:
    """
    Pre-multiplied BGRA to RGB. Alpha blending done with background.
    I.e. `RGB_color = mRGBA_color + (1 - alpha) x background`
    """
    b, g, r, a = cv2.split(bgra)
    if (ia := ~a).any():
        # rgb[:] += rgb_base * ia // 255
        for dst, lut in zip([r, g, b], _LUT_CA[rgb_base]):
            dst[:] += cv2.LUT(ia, lut)
    return cv2.merge([r, g, b])


# def _mbgra_to_rgb(bgra: np.ndarray, rgb_base: np.ndarray) -> np.ndarray:
#     """
#     Pre-multiplied BGRA to RGB. Only image colors are kept.
#     I.e. `RGB_color = alpha ? (mRGBA_color / alpha) : background`
#     """
#     bgra = cv2.cvtColor(bgra, cv2.COLOR_mRGBA2RGBA)  # alpha unscale
#     rgb, a = bgra[..., 2::-1], bgra[..., 3]
#     rgb[a == 0] = rgb_base  # fill fully transparent pixels
#     return rgb


@dataclass(frozen=True)
class _Item(Item):
    name: bytes
    osd: Openslide

    @property
    def key(self) -> str:
        return self.name.decode()

    def __array__(self) -> np.ndarray:
        bgra = np.empty((*self.shape[:2], 4), 'u1')
        OSD.openslide_read_associated_image(self.osd.ptr, self.name,
                                            c_void_p(bgra.ctypes.data))
        rgb = _mbgra_to_rgb(bgra, self.osd.bg_color)
        return np.ascontiguousarray(rgb)


@dataclass(frozen=True)
class _Lod(Lod):
    pool: int
    index: int
    osd: Openslide

    def crop(self, slices: tuple[slice, ...]) -> np.ndarray:
        (y_min2, y_max2), (x_min2, x_max2) = box_ \
            = [(s.start, s.stop) for s in slices]  # noqa: UP027
        box = np.array(box_)
        valid_box = box.T.clip([0, 0], self.shape[:2]).T

        (y_min, y_max), (x_min, x_max) = valid_box
        if y_min == y_max or x_min == x_max:  # Patch is outside slide
            return np.broadcast_to(self.osd.bg_color,
                                   (y_max2 - y_min2, x_max2 - x_min2, 3))

        bgra = np.empty((y_max - y_min, x_max - x_min, 4), dtype='u1')
        OSD.openslide_read_region(
            self.osd.ptr,
            c_void_p(bgra.ctypes.data),
            int(x_min * self.pool),
            int(y_min * self.pool),
            self.index,
            int(x_max - x_min),
            int(y_max - y_min),
        )

        rgb = _mbgra_to_rgb(bgra, self.osd.bg_color)

        offsets = np.abs(valid_box - box).ravel().tolist()
        if any(offsets):
            bg_color = self.osd.bg_color.tolist()
            return cv2.copyMakeBorder(
                rgb, *offsets, borderType=cv2.BORDER_CONSTANT, value=bg_color)

        return np.ascontiguousarray(rgb)


class Openslide(Driver):
    def __init__(self, path: Path):
        self.ptr = OSD.openslide_open(path.as_posix().encode())
        if not self.ptr:
            raise ValueError(f'File {path} cannot be opened')

        if err := OSD.openslide_get_error(self.ptr):
            raise ValueError(err)
        weakref.finalize(self, OSD.openslide_close, self.ptr)

        self._tags = self._get_metadata()

        bg_hex = self._tags.get('openslide.background-color', 'FFFFFF')
        self.bg_color: np.ndarray = np.frombuffer(bytes.fromhex(bg_hex), 'u1')

        self.spacing = None
        mpp = (self._tags.get(f'openslide.mpp-{ax}') for ax in 'yx')
        if s := [float(m) for m in mpp if m]:
            self.spacing = float(np.mean(s))

        # TODO: handle offsets
        # self.offset = *(int(self._tags.get(f'openslide.bounds-{ax}', 0))
        #                 for ax in 'yx'),
        # self.size = *(int(self._tags.get(f'openslide.bounds-{ax}', lim))
        #               for ax, lim in zip(('height', 'width'), self.shape)),

    def __repr__(self) -> str:
        return f'{type(self).__name__}({addressof(self.ptr.contents):0x})'

    def _get_metadata(self) -> dict[str, str]:
        names = OSD.openslide_get_property_names(self.ptr)
        return {
            name.decode(): value.decode() for name in _ntas_to_iter(names)
            if (value := OSD.openslide_get_property_value(self.ptr, name))
        }

    def __len__(self) -> int:
        return OSD.openslide_get_level_count(self.ptr)

    def __getitem__(self, index: int) -> Lod:
        h, w = c_int64(), c_int64()
        OSD.openslide_get_level_dimensions(self.ptr, index, byref(w), byref(h))
        pool = OSD.openslide_get_level_downsample(self.ptr, index)
        assert pool > 0

        return _Lod((h.value, w.value, 3), self.spacing, round(pool), index,
                    self)

    def keys(self) -> list[str]:
        names = OSD.openslide_get_associated_image_names(self.ptr)
        return [name.decode() for name in _ntas_to_iter(names)]

    def get(self, key: str) -> Item:
        name = key.encode()
        w, h = c_int64(), c_int64()
        OSD.openslide_get_associated_image_dimensions(self.ptr, name, byref(w),
                                                      byref(h))
        return _Item((h.value, w.value, 3), name, self)



"""
Driver based on libtiff
- fast
- not thread safe (internally)
- compatible with TIFF and its flavours
"""

from __future__ import annotations

__all__ = ['Tiff']

import ctypes
import sys
import weakref
from collections.abc import Iterator
from contextlib import contextmanager
from ctypes import (POINTER, addressof, byref, c_char_p, c_float, c_int,
                    c_ubyte, c_uint16, c_uint32, c_uint64, c_void_p,
                    create_string_buffer, string_at)
from dataclasses import dataclass, field
from enum import Enum
from pathlib import Path
from threading import Lock
from typing import NamedTuple, Protocol, TypeVar

import cv2
import imagecodecs
import numpy as np

from ._libs import load_library
from ._slide_bases import Driver, Item, Lod

_T = TypeVar('_T')

TIFF = load_library('libtiff', 6, 5)
# _TIFF.TIFFSetErrorHandler(None)

(TIFF.TIFFOpenW if sys.platform == 'win32' else TIFF.TIFFOpen).restype \
    = POINTER(c_ubyte)


# ---------------------------- TIFF tags mappings ----------------------------
class _Colorspace(Enum):
    MINISBLACK = 1
    RGB = 2
    # YCBCR = 6


class _Compression(Enum):
    RAW = 1
    CCITT = 2
    CCITTFAX3 = 3
    CCITTFAX4 = 4
    LZW = 5
    JPEG_OLD = 6
    JPEG = 7
    ADOBE_DEFLATE = 8
    RAW_16 = 32771
    PACKBITS = 32773
    THUNDERSCAN = 32809
    DEFLATE = 32946
    DCS = 32947
    JPEG2000_YUV = 33003
    JPEG2000_RGB = 33005
    JBIG = 34661
    SGILOG = 34676
    SGILOG24 = 34677
    JPEG2000 = 34712
    LZMA = 34925
    ZSTD = 50000
    WEBP = 50001


class _Tag:  # noqa: PIE795,RUF100
    JPEG_TABLES = 347
    TILE_BYTE_COUNTS = 325
    BACKGROUND_COLOR = 434


class _Tags:
    def __init__(self, ptr):
        self._ptr = ptr
        self.compression, = self._get(c_uint16, 259)
        self.colorspace, = self._get(c_uint16, 262)
        self.spp, = self._get(c_uint16, 277)
        self.is_planar, = self._get(c_uint16, 284)
        self.image_size = self._get(c_uint32, 257, 256)
        self.tile_size = self._get(c_uint32, 323, 322)
        dscr = self._get(c_char_p, 270)
        self.description: bytes = dscr and dscr.pop() or b''
        self.resolution = self._get(c_float, 283, 282)

    def _get(self, tp: type[ctypes._SimpleCData[_T]], *tags: int) -> list[_T]:
        values = []
        for tag in tags:
            cv = tp()
            if TIFF.TIFFGetField(self._ptr, c_uint32(tag), byref(cv)):
                values.append(cv.value)
        return values


# -------------------------- lazy decoding proxies ---------------------------
class SupportsArray(Protocol):
    def __array__(self) -> np.ndarray:
        raise NotImplementedError


class ImageArray(NamedTuple):
    data: object

    def __array__(self) -> np.ndarray:
        return imagecodecs.imread(self.data)


class JpegArray(NamedTuple):
    data: object
    jpt: bytes
    colorspace: int

    def __array__(self) -> np.ndarray:
        return imagecodecs.jpeg_decode(
            self.data, tables=self.jpt, colorspace=self.colorspace)


# -------------------- item, lod & opener implementations --------------------
@dataclass(frozen=True)
class _ItemBase(Item):
    index: int
    tiff: Tiff
    head: list[str]
    meta: dict[str, str]  # unused
    colorspace: _Colorspace
    bg_color: np.ndarray
    compression: _Compression
    jpt: bytes = field(repr=False)


@dataclass(frozen=True)
class _Item(_ItemBase):
    @property
    def key(self) -> str | None:
        if self.tiff.is_svs and self.index == 1:
            return 'thumbnail'
        for key in ('label', 'macro'):
            if any(key in s for s in self.head):
                return key
        return None

    def __array__(self) -> np.ndarray:
        h, w = self.shape[:2]
        bgra = np.empty((h, w, 4), dtype='u1')

        with self.tiff.ifd(self.index) as ptr:
            ok = TIFF.TIFFReadRGBAImage(ptr, w, h, c_void_p(bgra.ctypes.data),
                                        0)
            assert ok

        bgra = cv2.cvtColor(bgra, cv2.COLOR_mRGBA2RGBA)
        # TODO: do we need to use bg_color to fill points where alpha = 0 ?
        return bgra[..., 2::-1]


@dataclass(frozen=True)
class _Lod(Lod, _ItemBase):
    tile: tuple[int, ...]
    tile_sizes: np.ndarray = field(repr=False)

    def _get_tile(self, y, x, ptr) -> SupportsArray:
        offset = TIFF.TIFFComputeTile(ptr, x, y, 0, 0)
        nbytes = int(self.tile_sizes[offset])

        assert nbytes, 'File has corrupted tiles with zero size'
        if not nbytes:  # If nothing to read, don't read
            # TODO: read from previous lod
            # * If tile is empty on level N,
            # * then all tiles on levels >N are invalid, whether empty or not
            return np.broadcast_to(self.bg_color, self.tile)

        # if not self.compression.name.startswith('JPEG2000'):
        if self.compression not in {
                _Compression.JPEG2000_RGB, _Compression.JPEG2000_YUV
        }:
            image = np.empty(self.tile, dtype='u1')
            isok = TIFF.TIFFReadTile(ptr, c_void_p(image.ctypes.data), x, y, 0,
                                     0)
            assert isok != -1
            return image

        data = create_string_buffer(nbytes)
        TIFF.TIFFReadRawTile(ptr, offset, data, len(data))

        if self.jpt:
            return JpegArray(data, self.jpt, self.colorspace.value)
        return ImageArray(data)

    def crop(self, slices: tuple[slice, ...]) -> np.ndarray:
        box = np.array([(s.start, s.stop) for s in slices])

        *tile, spp = self.tile
        dy, dx = (low for low, _ in box)
        out = np.ascontiguousarray(
            np.broadcast_to(self.bg_color,
                            [hi - lo for lo, hi in box] + [spp]))

        hw = self.shape[:2]
        bmin, bmax = np.transpose(box).clip(0, hw)

        axes = *map(slice, bmin // tile, -(-bmax // tile)),
        t_lo = np.mgrid[axes].reshape(2, -1).T * tile  # [N, 2]
        if not t_lo.size:
            return out

        with self.tiff.ifd(self.index) as ptr:
            parts = [self._get_tile(y, x, ptr) for y, x in t_lo.tolist()]

        # [N, lo-hi, yx]
        crops = np.stack([t_lo, t_lo + tile], 1).clip(bmin, bmax)

        # [N, yx, lo-hi]
        o_crops = (crops - [dy, dx]).transpose(0, 2, 1)
        t_crops = (crops - t_lo[:, None, :]).transpose(0, 2, 1)
        for part, (oy, ox), (ty, tx) in zip(parts, o_crops, t_crops):
            patch = np.array(part, copy=False)
            out[slice(*oy), slice(*ox)] = patch[slice(*ty), slice(*tx)]

        return out


# FIXME: Get around slides from choked SVS encoder
class Tiff(Driver):
    def __init__(self, path: Path):
        # TODO: use memmap instead of libtiff
        spath = path.as_posix()
        self._ptr = (
            TIFF.TIFFOpenW(spath, b'rm') if sys.platform == 'win32' else
            TIFF.TIFFOpen(spath.encode(), b'rm'))
        if not self._ptr:
            raise ValueError(f'File {path} cannot be opened')

        weakref.finalize(self, TIFF.TIFFClose, self._ptr)

        self._lock = Lock()
        self.is_svs = path.suffix == '.svs'  # TODO: parse vendor tags for this

    def __repr__(self) -> str:
        return f'{type(self).__name__}({addressof(self._ptr.contents):0x})'

    @contextmanager
    def ifd(self, index: int) -> Iterator:
        with self._lock:
            TIFF.TIFFSetDirectory(self._ptr, index)
            try:
                yield self._ptr
            finally:
                TIFF.TIFFFreeDirectory(self._ptr, index)

    def __len__(self) -> int:
        return TIFF.TIFFNumberOfDirectories(self._ptr)

    def _bg_color(self) -> np.ndarray:
        bg_hex = b'FFFFFF'
        bg_color_ptr = c_char_p()
        if TIFF.TIFFGetField(self._ptr, _Tag.BACKGROUND_COLOR,
                             byref(bg_color_ptr)):
            bg_hex = string_at(bg_color_ptr, 3)
            # TIFF._TIFFfree(bg_color_ptr)  # TODO: ensure no segfault
        return np.frombuffer(bytes.fromhex(bg_hex.decode()), 'u1')

    def _parse_description(self,
                           desc: str) -> tuple[list[str], dict[str, str]]:
        head, *rest = desc.split('|')
        heads = [s.strip() for s in head.splitlines()]

        meta = {}
        for s in rest:
            if len(tags := s.split('=', 1)) == 2:
                meta[tags[0].strip()] = tags[1].strip()
            else:
                raise ValueError(f'Unparseable line in description: {s!r}')

        return heads, meta

    def _spacing(self, resolution: list[float],
                 meta: dict[str, str]) -> float | None:
        if s := [(10_000 / v) for v in resolution if v]:
            return float(np.mean(s))
        if mpp := meta.get('MPP'):
            return float(mpp)
        return None

    def _get(self, index: int) -> Item:
        tags = _Tags(self._ptr)

        if tags.is_planar != 1:
            raise TypeError(f'Level {index} is not contiguous!')

        bg_color = self._bg_color()
        head, meta = self._parse_description(tags.description.decode())
        spacing = self._spacing(tags.resolution, meta)

        colorspace = _Colorspace(tags.colorspace)

        # Compression and JPEG tables
        jpt = b''
        compression = _Compression(tags.compression)
        if compression is _Compression.JPEG:
            count = c_int()
            jpt_ptr = c_char_p()
            if TIFF.TIFFGetField(self._ptr, _Tag.JPEG_TABLES, byref(count),
                                 byref(jpt_ptr)) and count.value > 4:
                jpt = string_at(jpt_ptr, count.value)
                # TIFF._TIFFfree(jpt_ptr)  # TODO: ensure no segfault

        # Whole level shape
        shape = (*tags.image_size, tags.spp)
        if len(shape) != 3:
            raise ValueError(f'Bad shape in TIFF: {shape}')

        # Tile shape, if applicable
        if not TIFF.TIFFIsTiled(self._ptr):  # Not yet supported
            return _Item(shape, index, self, head, meta, colorspace, bg_color,
                         compression, jpt)

        # Tile sizes
        tile = (*tags.tile_size, tags.spp)
        if len(tile) != 3:
            raise ValueError(f'Bad tile shape in TIFF: {tile}')

        tbc = np.empty([], 'u8')
        tbc_ptr = POINTER(c_uint64)()
        if TIFF.TIFFGetField(self._ptr, _Tag.TILE_BYTE_COUNTS, byref(tbc_ptr)):
            num_tiles = TIFF.TIFFNumberOfTiles(self._ptr)
            tbc = np.ctypeslib.as_array(tbc_ptr, [num_tiles]).copy()
            # TIFF._TIFFfree(tbc_ptr)  # TODO: ensure no segfault
            if (tbc == 0).any():
                raise ValueError('Tiles are corrupted as their sizes have 0s')

        return _Lod(shape, index, self, head, meta, colorspace, bg_color,
                    compression, jpt, spacing, tile, tbc)

    def __getitem__(self, index: int) -> Item:
        with self.ifd(index):
            return self._get(index)



from __future__ import annotations

__all__ = ['Mosaic', 'Tile', 'get_fusion']

import dataclasses
from collections import defaultdict
from collections.abc import Callable, Iterable, Iterator
from contextlib import AbstractContextManager
from dataclasses import dataclass, field
from functools import partial
from itertools import chain
from typing import NamedTuple, Protocol, TypeVar, cast

import cv2
import numpy as np
from glow import chunked, map_n

from ._util import get_trapz

Vec = tuple[int, int]

# TODO: allow result of .map/.map_batched to have different tile and step

# ------------------------------- basic types --------------------------------


class NumpyLike(Protocol):
    @property
    def shape(self) -> tuple[int, ...]:
        ...

    def __getitem__(self, key: slice | tuple[slice, ...]) -> np.ndarray:
        ...


class Tile(NamedTuple):
    idx: Vec
    vec: Vec
    data: np.ndarray


# ---------------------------- utility functions -----------------------------


def _apply(fn: Callable[[np.ndarray], np.ndarray], obj: Tile) -> Tile:
    r = fn(obj.data)
    assert r.shape[:2] == obj.data.shape[:2], \
        'Tile shape alteration (besides channel count) is forbidden'
    return obj._replace(data=r)


def _apply_batched(fn: Callable[[list[np.ndarray]], Iterable[np.ndarray]],
                   ts: tuple[Tile, ...]) -> list[Tile]:
    *rs, = fn([t.data for t in ts])
    assert len(rs) == len(ts)
    assert all(r.shape[:2] == t.data.shape[:2]
               for r, t in zip(rs, ts)), \
        'Tile shape alteration (besides channel count) is forbidden'
    return [t._replace(data=r) for t, r in zip(ts, rs)]


def _reweight(weight: np.ndarray, tile: np.ndarray) -> np.ndarray:
    assert tile.dtype.kind == 'f'
    return np.einsum('hwc,h,w -> hwc', tile, weight, weight, optimize=True)


def _crop(tiles: Iterable[Tile], shape: tuple[int, ...]) -> Iterator[Tile]:
    h, w = shape
    for iyx, (y, x), tile in tiles:
        yield Tile(
            idx=iyx,
            vec=(y, x),
            data=tile[:h - y, :w - x],
        )


def get_fusion(tiles: Iterable[Tile],
               shape: tuple[int, ...] | None = None) -> np.ndarray | None:
    r: np.ndarray | None = None

    if shape is None:  # Collect all the tiles to compute destination size
        tiles = [*tiles]
        # N arrays of (yx + hw)
        yx_hw = np.array([[[t.vec, t.data.shape[:2]] for t in tiles]]).sum(1)
        # bottom left most edge
        shape = *yx_hw.max(0).tolist(),
    else:
        assert len(shape) == 2

    for _, (y, x), tile in tiles:
        if not tile.size:
            continue

        h, w, c = tile.shape
        if r is None:  # First iteration, initilize
            r = np.zeros((*shape, c), tile.dtype)

        assert c == r.shape[2]
        v = r[y:, x:][:h, :w]  # View to destination
        v[:] = tile[:v.shape[0], :v.shape[1]]  # Crop if needed

    return r


# ------------------------------ mosaic setting ------------------------------


@dataclass
class Mosaic:
    """
    Helper to split image to tiles and process them.

    Parameters:
    - step - Step between consecutive tiles
    - overlap - Count of pixels that will be shared among overlapping tiles

    So tile size = overlap + non-overlapped area + overlap = step + overlap,
    and non-overlapped area = step - overlap.
    """
    step: int
    overlap: int

    def __post_init__(self):
        assert 0 <= self.overlap <= self.step
        assert self.overlap % 2 == 0  # That may be optional

    def get_kernel(self) -> np.ndarray:
        return get_trapz(self.step, self.overlap)

    def iterate(self,
                image: NumpyLike,
                max_workers: int = 1) -> _TiledArrayView:
        """Read tiles from input image"""
        shape = image.shape[:2]
        ishape = *(len(range(0, s + self.overlap, self.step)) for s in shape),
        cells = np.ones(ishape, dtype=np.bool_)

        return _TiledArrayView(self, shape, cells, image, max_workers)


# --------------------------------- actions ----------------------------------

_Self = TypeVar('_Self', bound='_BaseView')


@dataclass
class _BaseView:
    m: Mosaic
    shape: tuple[int, ...]
    cells: np.ndarray

    @property
    def ishape(self) -> tuple[int, ...]:
        return self.cells.shape

    def __len__(self) -> int:
        return int(self.cells.sum())

    def report(self) -> dict[str, str]:
        """Cells and area usage"""
        used = int(self.cells.sum())
        total = self.cells.size
        coverage = (used / total) * (1 + self.m.overlap / self.m.step) ** 2
        return {'cells': f'{used}/{total}', 'coverage': f'{coverage:.0%}'}

    def __iter__(self) -> Iterator[Tile]:
        raise NotImplementedError

    def map(self: _Self,
            fn: Callable[[np.ndarray], np.ndarray],
            /,
            max_workers: int = 0) -> _Self:
        """
        Applies function to each tile.
        Note: change of tile shape besides channel count is forbidden.
        Each tile is HWC-ordered ndarray.
        Supports threading.
        """
        tile_fn = partial(_apply, fn)
        tiles = map_n(tile_fn, self, max_workers=max_workers)
        return cast(
            _Self,  # don't narrow type
            _IterView(self.m, self.shape, self.cells, tiles))

    def pool(self: _Self, stride: int = 1) -> _Self:
        """Resizes each tile to desired stride"""
        if stride == 1:
            return self
        shape = *(len(range(0, s, stride)) for s in self.shape),
        return cast(  # don't narrow type
            _Self,
            _DecimatedView(
                Mosaic(self.m.step // stride, self.m.overlap // stride), shape,
                self.cells, stride, self))

    def crop(self) -> _BaseView:
        return _IterView(self.m, self.shape, self.cells, _crop(
            self, self.shape))

    def zip_with(
            self, view: np.ndarray,
            v_scale: int) -> Iterator[tuple[Vec, Vec, np.ndarray, np.ndarray]]:
        """Extracts tiles from `view` simultaneously with tiles from self"""
        assert v_scale >= 1
        for tile in self:
            tw, th = tile.data.shape[:2]
            v = view[tile.vec[0] // v_scale:,
                     tile.vec[1] // v_scale:][:tw // v_scale, :th // v_scale]
            yield tile.idx, tile.vec, tile.data, v

    def with_cm(self: _Self, cm: AbstractContextManager) -> _Self:
        return cast(_Self, _CmView(self.m, self.shape, self.cells, self, cm))


@dataclass
class _View(_BaseView):
    def map_batched(self,
                    fn: Callable[[list[np.ndarray]], Iterable[np.ndarray]],
                    /,
                    batch_size: int = 1,
                    max_workers: int = 0) -> _View:
        """
        Applies function to batches of tiles.
        Note: change of tile shape besides channel count is forbidden.
        Each tile is HWC-ordered ndarray.
        Supports threading.
        """
        tile_fn = partial(_apply_batched, fn)

        chunks = chunked(self, batch_size)
        batches = map_n(tile_fn, chunks, max_workers=max_workers)
        tiles = chain.from_iterable(batches)

        return _IterView(self.m, self.shape, self.cells, tiles)

    def transform(self, fn: Callable[[Iterable[Tile]],
                                     Iterable[Tile]]) -> _View:
        """
        TODO: fill docs
        """
        tiles = fn(self)
        return _IterView(self.m, self.shape, self.cells, tiles)

    def reweight(self) -> _View:
        """
        Applies weight to tile edges to prepare them for summation
        if overlap exists.
        Note: no need to call this method if you have already applied it.
        """
        if not self.m.overlap:  # No need
            return self

        weight = self.m.get_kernel()
        tile_fn = partial(_reweight, weight)
        return self.map(tile_fn)

    def merge(self) -> _BaseView:
        """
        Removes overlapping regions from all the tiles if any.
        Tiles should be reweighted if overlap exists.
        """
        if self.m.overlap:
            return _UniqueTileView(self.m, self.shape, self.cells, self)
        return self


@dataclass
class _IterView(_View):
    source: Iterable[Tile]

    def __iter__(self) -> Iterator[Tile]:
        return iter(self.source)


@dataclass
class _CmView(_BaseView):
    source: Iterable[Tile]
    _cm: AbstractContextManager

    def __iter__(self) -> Iterator[Tile]:
        with self._cm:
            yield from self.source


@dataclass
class _DecimatedView(_View):
    """
    Decimates tiles.
    Doesn't change size uniformity.
    Yields decimated views of original tiles.
    """
    stride: int
    source: Iterable[Tile]

    def __iter__(self) -> Iterator[Tile]:
        for t in self.source:
            vec = *(c // self.stride for c in t.vec),
            yield Tile(
                idx=t.idx,
                vec=vec,  # type: ignore[arg-type]
                data=t.data[::self.stride, ::self.stride],
            )


@dataclass
class _TiledArrayView(_View):
    """
    Extracts tiles from array.
    Yields same-sized tiles with overlaps.
    """
    data: NumpyLike
    max_workers: int

    def select(self, mask: np.ndarray, scale: int) -> _TiledArrayView:
        """Drop tiles where `mask` is 0"""
        assert mask.ndim == 2
        mask = mask.astype('u1')

        ih, iw = self.ishape
        step = self.m.step // scale
        pad = self.m.overlap // (scale * 2)

        mh, mw = (ih * step), (iw * step)
        if mask.shape[:2] != (mh, mw):
            mask_pad = [(0, s1 - s0) for s0, s1 in zip(mask.shape, (mh, mw))]
            mask = np.pad(mask, mask_pad)[:mh, :mw]

        if self.m.overlap:
            kernel = np.ones((3, 3), dtype='u1')
            mask = cv2.dilate(mask, kernel, iterations=pad)

        if pad:
            mask = np.pad(mask[:-pad, :-pad], [[pad, 0], [pad, 0]])

        cells = mask.reshape(ih, step, iw, step).any((1, 3))
        return dataclasses.replace(self, cells=cells)

    def _get_tile(self, iy: int, ix: int) -> Tile:
        """Read non-overlapping tile of source image"""
        (y0, y1), (x0, x1) = ((
            self.m.step * i - self.m.overlap,
            self.m.step * (i + 1),
        ) for i in (iy, ix))
        if iy and self.cells[iy - 1, ix]:
            y0 += self.m.overlap
        if ix and self.cells[iy, ix - 1]:
            x0 += self.m.overlap
        return Tile(
            idx=(iy, ix),
            vec=(y0, x0),
            data=self.data[y0:y1, x0:x1],
        )

    def _rejoin_tiles(self, image_parts: Iterable[Tile]) -> Iterator[Tile]:
        """Joins non-overlapping parts to tiles"""
        assert self.m.overlap
        overlap = self.m.overlap
        cells = np.pad(self.cells, [(0, 1), (0, 1)])
        step = self.m.step
        row = defaultdict[int, np.ndarray]()

        for (iy, ix), _, part in image_parts:
            # Lazy init, first part is always whole
            if row.default_factory is None:
                row.default_factory = partial(np.zeros, part.shape, part.dtype)

            if (tile := row.pop(ix, None)) is not None:
                tile[-part.shape[0]:, -part.shape[1]:] = part
            else:
                tile = part

            yield Tile(
                idx=(iy, ix),
                vec=(iy * step - overlap, ix * step - overlap),
                data=tile,
            )

            if cells[iy, ix + 1]:
                row[ix + 1][:, :overlap] = tile[:, -overlap:]
            if cells[iy + 1, ix]:
                row[ix][:overlap, :] = tile[-overlap:, :]

    def __iter__(self) -> Iterator[Tile]:
        """
        Yield complete tiles built from source image.
        Each tile will have size `(step + overlap)`
        """
        ys, xs = np.where(self.cells)
        parts = map_n(self._get_tile, ys, xs, max_workers=self.max_workers)
        return self._rejoin_tiles(parts) if self.m.overlap else iter(parts)


@dataclass
class _UniqueTileView(_BaseView):
    """
    Applies weighted average over overlapping regions.
    Yields tiles without overlaps, so their size can differ.
    """
    source: Iterable[Tile]

    _cells: np.ndarray = field(init=False, repr=False)
    _row: dict[int, np.ndarray] = field(init=False, repr=False)
    _carry: list[np.ndarray] = field(init=False, repr=False)

    def __post_init__(self):
        self._cells = np.pad(self.cells, [(0, 1), (0, 1)])
        self._row = {}
        self._carry = []

    def _update(self, obj: Tile) -> Tile:
        """
        Blends edges of overlapping tiles and returns non-overlapping parts
        """
        (iy, ix), (y, x), tile = obj
        overlap = self.m.overlap
        step = self.m.step

        if iy and self._cells[iy - 1, ix]:  # TOP exists
            top = self._row.pop(ix)
            tile[:overlap, step - top.shape[1]:step] += top
        else:
            tile = tile[overlap:]  # cut TOP
            y += overlap

        if ix and self._cells[iy, ix - 1]:  # LEFT exists
            left = self._carry.pop()
            if self._cells[iy + 1, [ix - 1, ix]].all():
                tile[-left.shape[0]:, :overlap] += left
            else:  # cut BOTTOM-LEFT
                tile[-left.shape[0] - overlap:-overlap, :overlap] += left
        else:
            tile = tile[:, overlap:]  # cut LEFT
            x += overlap

        tile, right = np.split(tile, [-overlap], axis=1)
        if self._cells[iy, ix + 1]:  # RIGHT exists
            if not (iy and self._cells[iy - 1, [ix, ix + 1]].all()):
                right = right[-step:]  # cut TOP-RIGHT
            if not self._cells[iy + 1, [ix, ix + 1]].all():
                right = right[:-overlap]  # cut BOTTOM-RIGHT
            self._carry.append(right)

        tile, bottom = np.split(tile, [-overlap])
        if self._cells[iy + 1, ix]:  # BOTTOM exists
            if not (ix and self._cells[[iy, iy + 1], ix - 1].all()):
                # cut BOTTOM-LEFT
                bottom = bottom[:, -(step - overlap):]
            self._row[ix] = bottom

        return Tile(
            idx=(iy, ix),
            vec=(y, x),
            data=tile,
        )

    def __iter__(self) -> Iterator[Tile]:
        assert self.m.overlap
        return map(self._update, self.source)



__all__ = ['get_trapz', 'probs_to_hsv']

import cv2
import numpy as np


def probs_to_hsv(prob: np.ndarray) -> np.ndarray:
    """Converts probability array of (H W C) shape to (H W 3) RGB image"""
    assert prob.ndim == 3
    if not prob.size:
        return np.empty((*prob.shape[:2], 3), dtype='u1')

    hue = prob.argmax(-1).astype('u1')
    value = prob.max(-1)
    if value.dtype != 'u1':
        value *= 255
        value = value.round().astype('u1')

    hue_lut = (np.arange(256) * (127 / prob.shape[2])).round().astype('u1')
    hsv = cv2.merge([
        cv2.LUT(hue, hue_lut),
        np.broadcast_to(np.uint8(255), prob.shape[:2]),
        value,
    ])
    return cv2.cvtColor(hsv, cv2.COLOR_HSV2RGB)


def get_trapz(step: int, overlap: int) -> np.ndarray:
    """Returns trapezional window kernel to apply stiching"""
    assert overlap
    pad = np.linspace(0, 1, overlap + 2)[1:-1]  # strip initial 0 and final 1
    return np.r_[pad, np.ones(step - overlap), pad[::-1]].astype('f4')
